#!/usr/bin/env python3
import plotly.express as px
from pandas import DataFrame

# df = pd.DataFrame(data=dict(x=[1, 2, 3, 4], y=[1, 4, 9, 16]))
# fig = px.line(df, x="x", y="y", title="test")
# fig.show()

class Line:
    def __init__(self, from_, to):
        self.from_ = from_
        self.to = to

    def slope(self) -> float:
        return (self.to[1] - self.from_[1]) / (self.to[0] - self.from_[0])

    def get(self, x: float) -> float:
        from_x, from_y = self.from_
        to_x, to_y = self.to
        if x == from_x:
            return from_y
        tween = (x - from_x) / (to_x - from_x)
        return tween * to_y + (1-tween) * from_y


class Piecewise:
    def __init__(self, xy: list):
        """ xy is a list of (x, y) pairs """
        self.xy = list(xy)

    def normalized(self, prev_max: float) -> 'Piecewise':
        """ map every coordinate from [-prev_max, prev_max] to [0, 1] """
        p = prev_max
        r = 2*prev_max
        s = 1.0/r
        return Piecewise([
            (s*(x + p), s*(y + p)) for (x, y) in self.xy
        ])

    def logically_inverted(self) -> 'Piecewise':
        """ return a Piecewise that evaluates to 1-y """
        return Piecewise([
            (x, 1-y) for (x, y) in self.xy
        ])

    def logically_inverted_x(self) -> 'Piecewise':
        return Piecewise([
            (1-x, y) for (x, y) in self.xy
        ][::-1])

    def logically_inverted_xy(self) -> 'Piecewise':
        return self.logically_inverted_x().logically_inverted()

    def shifted_x(self, shift: float) -> 'Piecewise':
        return Piecewise([
            (x + shift, y) for (x, y) in self.xy
        ])

    def shifted_y(self, shift: float) -> 'Piecewise':
        return Piecewise([
            (x, y + shift) for (x, y) in self.xy
        ])

    def line_for(self, x: float) -> Line:
        for first_lower in self.xy[:-1][::-1]:
            if first_lower[0] < x: break
        for first_upper in self.xy[1:]:
            if not first_upper[0] < x: break

        return Line(first_lower, first_upper)

    def get(self, x: float) -> float:
        """
        evaluate the piecewise function at the provided x value.
        OOB points are just extrapolated from the nearest piece.
        """
        return self.line_for(x).get(x)

    def get_slope(self, x: float) -> float:
        return self.line_for(x).slope()

    def get_repeated(self, x: float, n: int = 255) -> float:
        for _ in range(n):
            x = self.get(x)
        return x

    def get_mean(self, x: float, n: int = 100) -> float:
        return 1/n * sum(self.get(x * i/(n-1)) for i in range(n))

    def get_integral(self, x: float, n: int = 100) -> float:
        return self.get_mean(x) * x

    def df_for(self, from_: float, to: float, points: int, f) -> DataFrame:
        x_step = (to - from_) / (points - 1)
        x = [from_ + x_step*x for x in range(points)]
        y = [f(xi) for xi in x]
        return DataFrame(data=dict(x=x, y=y))

    def df(self, from_: float = 0.0, to: float = 1.0, points: int = 101) -> DataFrame:
        return self.df_for(from_, to, points, self.get)

    def slope_df(self, from_: float = 0.0, to: float = 1.0, points: int = 101) -> DataFrame:
        return self.df_for(from_, to, points, self.get_slope)

    def plot_for(self, from_: float, to: float, title: str, f):
        df = self.df_for(from_, to, points=101, f=f)
        fig = px.line(df, x="x", y="y", title=title)
        fig.show()

    def plot(self, from_: float = 0.0, to: float = 1.0, title: str = "Piecewise"):
        self.plot_for(from_, to, title, self.get)

    def plot_slope(self, from_: float = 0.0, to: float = 1.0, title: str = "Piecewise"):
        self.plot_for(from_, to, title, self.get_slope)

    def plot_equilibrium(self, from_: float = 0.0, to: float = 1.0, title: str = "Piecewise"):
        self.plot_for(from_, to, title, self.get_repeated)

    def plot_integral(self, from_: float = 0.0, to: float = 1.0, title: str = "Piecewise"):
        self.plot_for(from_, to, title, self.get_integral)


# stable inverter (ideal)
fwd_fake_step = Piecewise(
    [
        [ 0.0, 0.0 ],
        [ 0.4, 0.0 ],
        [ 0.6, 1.0 ],
        [ 1.0, 1.0 ],
    ]
)

# stable inverter (amplifying)
fwd_fake_1_5x = Piecewise(
    [
        [ 0.0,  0.0 ],
        [ 0.65, 1.0 ],
        [ 1.0,  1.0 ],
    ]
)
# stable inverter (amplifying only from 0.3 -> 0.5)
fwd_fake_slope_change_before_0_5 = Piecewise(
    [
        [ 0.0, 0.2 ],
        [ 0.3, 0.3 ],
        [ 0.5, 0.6 ],
        [ 1.0, 1.0 ],
    ]
)
# failed inverter (>1.0 slope happens too late)
# flipping x doesn't fix.
# however, shifting x by -0.1 and y by -0.2 and *then* inverting x does.
# - this gives us a concave-up 1-x like curve
fwd_fake_slope_change_after_0_5 = Piecewise(
    [
        [ 0.0, 0.2 ],
        [ 0.3, 0.3 ],
        [ 0.6, 0.5 ],
        [ 1.0, 1.0 ],
    ]
)

slope_fake_hill = [
  0.8, 0.9, 1.0, 1.1, 1.2, 1.2, 1.1, 1.0, 0.9, 0.8
]
fwd_fake_hill = Piecewise(
    [ (0.1*i, 0.1 * sum(slope_fake_hill[0:i])) for i in range(11) ]
)

fwd_fake_asymmetric_hill = Piecewise(
    [
        (0.0, 0.20),
        (0.2, 0.30),
        (0.4, 0.45),
        (0.6, 0.75),
        (0.8, 0.80),
        (1.0, 0.85),
    ]
)
# valid inverter; the [0.6, 1.0] -> 0.8 mapping *fixes* the logic low value to
#   1.0 - 0.8 = 0.2
#   and allows anything 0.6 to 1.0 to be recognized as logic high immediately.
# i.e. "bottoming out" is a *good* thing
fwd_fake_asymmetric_flats = Piecewise(
    [
        (0.0, 0.20),
        (0.2, 0.30),
        (0.6, 0.80),
        (1.0, 0.80),
    ]
)

fwd_fake_asymmetric_overdrive = Piecewise(
    [
        (0.0, 0.40),
        (0.3, 0.50),
        (0.6, 0.85),
        (1.0, 0.90),
    ]
)

fwd_fake_asymmetric_bottom_out = Piecewise(
    [
        (0.0, 0.00),
        (0.8, 0.99),
        (1.0, 1.00),
    ]
)


fwd_17_4_0_8e10 = Piecewise(
    [
        [ -16381,   6688 ],
        [ -15885,   6778 ],
        [ -14831,   6878 ],
        [ -13622,   7004 ],
        [   -883,   8528 ],
        [   6252,   9496 ],
        [   7846,   9703 ],
        [   8148,   9766 ],
        [   8425,   9831 ],
        [   8705,   9892 ],
        [   8988,   9916 ],
        [   9866,  10114 ],
        [  11179,  10234 ],
        [  12033,  10382 ],
        [  12491,  10422 ],
        [  13135,  10494 ],
        [  14363,  10649 ],
    ]
).normalized(17000)

fwd_18 = Piecewise(
    [
        [ -16206, -1131 ],
        [ -15192,  -746 ],
        [ -12827,    33 ],
        [   -642,  4990 ],
        [  13082,  9652 ],
        [  16696, 10600 ],
    ]
).normalized(17000)


fwd_24_5_1_2e10 = Piecewise(
    [
        [ -12912,  -8487 ],
        [  -4754,  -6045 ],
        [   2687,  -2560 ],
        [   3936,  -1774 ],
        [   4267,  -1517 ],
        [   4504,  -1314 ],
        [   4710,  -1132 ],
        [   4820,  -1075 ],
        [   4884,  -1042 ],
        [   4948,  -1012 ],
        [   5046,   -968 ],
        [   5205,   -897 ],
        [   5364,   -829 ],
        [   5525,   -760 ],
        [   5843,   -622 ],
        [   6764,   -197 ],
        [   9467,    788 ],
    ]
).normalized(15000)

fwd_24_5_1_5e10 = Piecewise(
    [
        [ -15208,  -6303 ],
        [ -13396,  -5388 ],
        [ -11992,  -4516 ],
        [ -11991,  -4499 ],
        [  -9379,  -2953 ],
        [  -4757,    531 ],
        [     -2,   4734 ],
        [   3074,   7760 ],
        [   4854,   9784 ],
        [   5611,  10736 ],
        [   5994,  11126 ],
        [   6298,  11404 ],
        [   6678,  11757 ],
        [   7196,  12200 ],
        [   7667,  12589 ],
        [   8238,  13048 ],
        [   8239,  13046 ],
        [   9613,  14027 ],
        [  10585,  14622 ],
        [  12048,  15346 ],
    ]
).normalized(17000)

fwd_24_5_1_8e10 = Piecewise(
    [
        [ -16412,  -3392 ],
        [ -15266,  -2681 ],
        [ -14036,  -1897 ],
        [ -12789,  -1110 ],
        [  -8766,   1588 ],
        [  -2052,   6544 ],
        [   2389,   9989 ],
        [   4225,  11437 ],
        [   5194,  12182 ],
        [   5971,  12438 ],
        [   6901,  12937 ],
        [   8308,  13632 ],
        [   9910,  14365 ],
        [  10583,  14662 ],
        [  11240,  14850 ],
        [  12114,  15171 ],
        [  13862,  15600 ],
    ]
).normalized(17000)

fwd_26 = Piecewise(
    [
        [ -14687, -7326 ],
        [ -13049, -6503 ],
        [ -11785, -5833 ],
        [ -4649,  -1447 ],
        [ 4961,    7059 ],
        [ 11283,  11147 ],
    ]
).normalized(17000)


fwd_38_1_0 = Piecewise(
    [
        [ -12817, -8131 ],  # -1.00
        [ -12239, -7798 ],  # -0.80
        [ -4859,  -2587 ],  # -0.50
        [ 1490,    3012 ],  # -0.30
        [ 3866,    5327 ],  # -0.20
        [ 6030,    7237 ],  # -0.10
        [ 7747,    8357 ],  #  0.00
        [ 9494,    9202 ],  # +0.10
        [ 11261,  10011 ],  # +0.20
        [ 12941,  10808 ],  # +0.30
        [ 15415,  11986 ],  # +0.50
        [ 16196,  12375 ],  # +0.80
        # [ 16182,  12352 ],  # +1.00
    ]
).normalized(17000)

fwd_38_1_0_5e10 = Piecewise(
    [
        [ -16180, -7079 ],  # -1.00
        [ -14443, -5965 ],  # -0.50
        [ -5579,    -13 ],  # -0.20
        [ 10033,   7676 ],  #  0.00
        [ 14986,   9375 ],  # +0.20
        [ 15149,   9606 ],  # +0.50
        [ 15801,   9924 ],  # +1.00
    ]
).normalized(17000)

fwd_38_2_0 = Piecewise(
    [
        [ (-13745 + -13012)/2, -6222 ],  # -1.00
        [ (-13097 + -12338)/2, -5662 ],  # -0.80
        [ (-4969 + -4744)/2,    2373 ],  # -0.50
        [ (535 + 611)/2,        8793 ],  # -0.30
        [ (1772 + 2070)/2,     10467 ],  # -0.20
        [ (3143 + 3200)/2,     11906 ],  # -0.10
        [ (4472 + 4114)/2,     12921 ],  #  0.00
        [ (5838 + 5144)/2,     13788 ],  # +0.10
        [ (7221 + 6291)/2,     14530 ],  # +0.20
        [ (8558 + 7644)/2,     15127 ],  # +0.30
        [ (11159 + 10397)/2,   15865 ],  # +0.50
        [ (12778 + 14243)/2,   16162 ],  # +0.80
        [ (12430 + 15653)/2,   16202 ],  # +1.00
    ]
).normalized(17000)


fwd_38_2_0_5e10 = Piecewise(
    [
        [ (-16386 + -16170)/2, -3490 ],  # -1.00
        [ (-16107 + -15529)/2, -3035 ],  # -0.80
        [ (-15075 + -14122)/2, -1827 ],  # -0.50
        [ (-13387 + -12396)/2,   -63 ],  # -0.30
        [ (-5358 + -5201)/2,    7423 ],  # -0.20
        [ (2355 + 1719)/2,     12039 ],  # -0.10
        [ (7563 + 5962)/2,     13479 ],  #  0.00
        [ (10617 + 9318)/2,    14282 ],  # +0.10
        [ (12779 + 12447)/2,   14796 ],  # +0.20
        [ (12649 + 15269)/2,   15034 ],  # +0.30
        [ (13077 + 16320)/2,   15140 ],  # +0.50
        [ (14410 + 16557)/2,   15260 ],  # +0.80
        [ (15281 + 16623)/2,   15331 ],  # +1.00
    ]
).normalized(17000)

fwd_38_3_0 = Piecewise(
    [
        [ (-13956 + -13890 + -13077)/3, -5203 ],  # -1.00
        [ (-13292 + -13161 + -12374)/3, -4518 ],  # -0.80
        [ (-4979 + -4885  + -4717)/3,    5051 ],  # -0.50
        [ (381 + -153 + 31)/3,          11264 ],  # -0.30
        [ (1531 + 503 + 1006)/3,        12509 ],  # -0.20
        [ (2862 + 1120 + 1743)/3,       13549 ],  # -0.10
        [ (4180 + 1821 + 2239)/3,       14386 ],  #  0.00
        [ (5560 + 2564 + 2899)/3,       15033 ],  # +0.10
        [ (6986 + 3436 + 3701)/3,       15451 ],  # +0.20
        [ (8358 + 4396 + 4732)/3,       15738 ],  # +0.30
        [ (10482 + 6644 + 7735)/3,      16081 ],  # +0.50
        [ (11246 + 12478 + 12663)/3,    16343 ],  # +0.80
        [ (11436 + 13343 + 14411)/3,    16380 ],  # +1.00
    ]
).normalized(17000)

fwd_38_3_0_5e10 = Piecewise(
    [
        [ (-16403 + -16389 + -16152)/3, -1175 ],  # -1.00
        [ (-16134 + -16084 + -15471)/3,  -701 ],  # -0.80
        [ (-15192 + -14891 + -14016)/3,   777 ],  # -0.50
        [ (-13512 + -13089 + -12278)/3,  2939 ],  # -0.30
        [ (-5248 + -5187 + -5032)/3,    11125 ],  # -0.20
        [ (2099 + 645 + 708)/3,         14046 ],  # -0.10
        [ (7045 + 3536 + 3757)/3,       14557 ],  #  0.00
        [ (9729 + 6054 + 6543)/3,       14967 ],  # +0.10
        [ (11453 + 9238 + 10081)/3,     15393 ],  # +0.20
        [ (11572 + 13274 + 13839)/3,    15759 ],  # +0.30
        [ (12534 + 15192 + 16090)/3,    15925 ],  # +0.50
        [ (14013 + 16353 + 16508)/3,    16007 ],  # +0.80
        [ (14944 + 16565 + 16606)/3,    16033 ],  # +1.00
    ]
).normalized(17000)

fwd_38_4_0 = Piecewise(
    [
        [ (-14020 + -14112 + -13935 + -13091)/4, -4701 ],  # -1.00
        [ (-13353 + -13363 + -13185 + -12381)/4, -3947 ],  # -0.80
        [ (-4982 + -4912 + -4870 + -4696)/4,      6398 ],  # -0.50
        [ (338 + -243 + -352 + -254)/4,          12205 ],  # -0.30
        [ (1469 + 303 + 107 + 510)/4,            13165 ],  # -0.20
        [ (2789 + 839 + 443 + 1089)/4,           13989 ],  # -0.10
        [ (4150 + 1560 + 653 + 1416)/4,          14727 ],  #  0.00
        [ (5562 + 2421 + 979 + 1899)/4,          15224 ],  # +0.10
        [ (7027 + 3336 + 1460 + 2518)/4,         15551 ],  # +0.20
        [ (8402 + 4357 + 2093 + 3293)/4,         15802 ],  # +0.30
        [ (10385 + 7210 + 5554 + 5998)/4,        16123 ],  # +0.50
        [ (11301 + 12558 + 13004 + 12508)/4,     16388 ],  # +0.80
        [ (11462 + 13429 + 13829 + 13370)/4,     16408 ],  # +1.00
    ]
).normalized(17000)

fwd_38_4_0_5e10 = Piecewise(
    [
        [ (-16395 + -16418 + -16377 + -16134)/4,  346 ],  # -1.00
        [ (-16128 + -16140 + -16066 + -15431)/4,  843 ],  # -0.80
        [ (-15217 + -15075 + -14834 + -13958)/4, 2407 ],  # -0.50
        [ (-13556 + -13263 + -13009 + -12226)/4, 4772 ],  # -0.30
        [ (-5186 + -5134 + -5083 + -4922)/4,    12813 ],  # -0.20
        [ (1894 + 537 + -149 + 331)/4,          14556 ],  # -0.10
        [ (6845 + 3237 + 1699 + 2725)/4,        14936 ],  # 0.00
        [ (9591 + 5649 + 3599 + 5023)/4,        15250 ],  # +0.10
        [ (11255 + 8605 + 7601 + 8571)/4,       15647 ],  # +0.20
        [ (11268 + 12837 + 13307 + 13021)/4,    15944 ],  # +0.30
        [ (12388 + 14744 + 15209 + 15530)/4,    16104 ],  # +0.50
        [ (13871 + 16190 + 16345 + 16412)/4,    16178 ],  # +0.80
        [ (14774 + 16497 + 16563 + 16559)/4,    16197 ],  # +1.00
    ]
).normalized(17000)

inv_39_2_0_2e10 = Piecewise(
    [
        [ -12902,  10759 ],
        [ -12339,  11336 ],
        [  -8581,  11274 ],
        [  -4821,  10571 ],
        [   -822,   9463 ],
        [   3117,   8265 ],
        [   4938,   7704 ],
        [   6441,   7221 ],
        [   7234,   6912 ],
        [   7844,   6662 ],
        [   8282,   6551 ],
        [   8674,   6443 ],
        [   9071,   6325 ],
        [   9479,   6191 ],
        [  10311,   5885 ],
        [  11153,   5541 ],
        [  12833,   4788 ],
        [  14097,   4071 ],
        [  14561,   3816 ],
    ]
).normalized(15000)

inv_39_2_0_5e10 = Piecewise(
    [
        [ -15691,   9609 ],
        [ -15154,   9450 ],
        [ -14498,   9327 ],
        [ -14086,   9217 ],
        [ -13501,   9113 ],
        [ -12664,   8923 ],
        [  -9677,   7937 ],
        [  -4868,   5948 ],
        [    222,   3390 ],
        [   5223,    610 ],
        [   9175,  -1732 ],
        [  11286,  -2820 ],
        [  12505,  -3439 ],
        [  13504,  -3957 ],
        [  14679,  -4588 ],
        [  15127,  -4830 ],
        [  15667,  -5110 ],
        [  16156,  -5353 ],
        [  16350,  -5450 ],
    ]
).normalized(17000)

inv_39_2_0_8e10 = Piecewise(
    [
        [ -16465,   6854 ],
        [ -16318,   6905 ],
        [ -16079,   6824 ],
        [ -15789,   6623 ],
        [ -15296,   6435 ],
        [ -14593,   6212 ],
        [ -14052,   5989 ],
        [ -13259,   5563 ],
        [  -8825,   3572 ],
        [  -1149,   -218 ],
        [   6004,  -3851 ],
        [  10704,  -6032 ],
        [  13131,  -6986 ],
        [  14268,  -7421 ],
        [  14894,  -7683 ],
        [  15301,  -7829 ],
        [  15839,  -8028 ],
        [  16356,  -8243 ],
        [  16507,  -8292 ],
    ]
).normalized(17000)


inv_39_2_0_1e11 = Piecewise(
    [
        [ -16651,   5123 ],
        [ -16567,   5111 ],
        [ -16429,   5092 ],
        [ -16312,   5120 ],
        [ -16102,   5078 ],
        [ -15572,   4837 ],
        [ -15109,   4545 ],
        [ -14393,   4298 ],
        [ -13129,   3647 ],
        [  -5324,   -167 ],
        [   3762,  -4392 ],
        [  10248,  -7171 ],
        [  13522,  -8341 ],
        [  14221,  -8595 ],
        [  14851,  -8807 ],
        [  15280,  -8964 ],
        [  15864,  -9130 ],
        [  16440,  -9363 ],
        [  16585,  -9409 ],
    ]
).normalized(17000)

inv_39_2_0_15e10 = Piecewise(
    [
        [ -16854,   1899 ],
        [ -16811,   1926 ],
        [ -16759,   1908 ],
        [ -16723,   1910 ],
        [ -16670,   1908 ],
        [ -16569,   1907 ],
        [ -16466,   1877 ],
        [ -16269,   1775 ],
        [ -15731,   1520 ],
        [ -13797,    601 ],
        [  -1756,  -4314 ],
        [   9395,  -7804 ],
        [  13461,  -8670 ],
        [  14026,  -8763 ],
        [  14766,  -8878 ],
        [  15279,  -9140 ],
        [  16084,  -9300 ],
        [  16568,  -9413 ],
        [  16672,  -9386 ],
    ]
).normalized(17000)


# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.0004rad-12coupling-3_1_winding-5e10-drive-
fwd_40_12_3_1_5e10 = Piecewise(
    [
        [ -16453,  -5877 ],  # -2.0
        [ -16171,  -5444 ],  # -1.5
        [ -15617,  -4900 ],  # -1.0
        [ -15086,  -4405 ],  # -0.8
        [ -14357,  -3941 ],  # -0.6
        [ -13828,  -3800 ],  # -0.5
        [ -13111,  -3294 ],  # -0.4
        [ -12203,  -2582 ],  # -0.3
        [  -9321,   -169 ],  # -0.25
        [  -4645,   4342 ],  # -0.2
        [    422,   9543 ],  # -0.15
        [   5472,  14584 ],  # -0.1
        [   8233,  16078 ],  # -0.05
        [   9249,  16224 ],  # 0.0
        [   9700,  16263 ],  # 0.05
        [  10044,  16286 ],  # 0.1
        [  10643,  16323 ],  # 0.2
        [  11168,  16348 ],  # 0.3
        [  11605,  16367 ],  # 0.5
        [  12054,  16387 ],  # 0.8
        [  12411,  16402 ],  # 1.0
        [  13301,  16442 ],  # 1.5
        [  14146,  16474 ],  # 2.0
    ]
).normalized(17000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.0004rad-6coupling-3_1_winding-5e10-drive-
fwd_40_6_3_1_5e10 = Piecewise(
    [
        [ -16689,  -6116 ],  # -2.0
        [ -16510,  -5954 ],  # -1.5
        [ -16046,  -5674 ],  # -1.0
        [ -15523,  -5367 ],  # -0.8
        [ -14807,  -4923 ],  # -0.6
        [ -14350,  -4632 ],  # -0.5
        [ -13668,  -4326 ],  # -0.4
        [ -12591,  -3719 ],  # -0.3
        [  -9752,  -1873 ],  # -0.25
        [  -5310,   1484 ],  # -0.2
        [  -1048,   4991 ],  # -0.15
        [   2517,   7918 ],  # -0.1
        [   5284,  10134 ],  # -0.05
        [   7297,  11623 ],  # 0.0
        [   8629,  12439 ],  # 0.05
        [   9611,  12955 ],  # 0.1
        [  11360,  13845 ],  # 0.2
        [  12188,  14174 ],  # 0.3
        [  12831,  14537 ],  # 0.5
        [  13573,  14838 ],  # 0.8
        [  14082,  14999 ],  # 1.0
        [  15309,  15457 ],  # 1.5
        [  16329,  15683 ],  # 2.0
    ]
).normalized(17000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.0004rad-6coupling-7_1_winding-5e10-drive-
fwd_40_6_7_1_5e10 = Piecewise(
    [
        [ -16244,  -8159 ],  # -2.0
        [ -15756,  -7835 ],  # -1.5
        [ -14687,  -7326 ],  # -1.0
        [ -14108,  -7030 ],  # -0.8
        [ -13435,  -6690 ],  # -0.6
        [ -13049,  -6503 ],  # -0.5
        [ -12522,  -6223 ],  # -0.4
        [ -11785,  -5833 ],  # -0.3
        [  -9190,  -4410 ],  # -0.25
        [  -4649,  -1447 ],  # -0.2
        [    250,   2263 ],  # -0.15
        [   3132,   4799 ],  # -0.1
        [   4306,   6177 ],  # -0.05
        [   4961,   7059 ],  # 0.0
        [   5320,   7400 ],  # 0.05
        [   5631,   7670 ],  # 0.1
        [   6516,   8370 ],  # 0.2
        [   8605,   9674 ],  # 0.3
        [   9764,  10342 ],  # 0.5
        [  10753,  10874 ],  # 0.8
        [  11283,  11147 ],  # 1.0
        [  12329,  11657 ],  # 1.5
        [  13223,  12045 ],  # 2.0
    ]
).normalized(15000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-20coupling-3_1_winding-3e10-drive-
fwd_40_600um_20_3_1_3e10 = Piecewise(
    [
        [  -9940,  -2886 ],  # -2.0
        [  -9292,  -2432 ],  # -1.5
        [  -8364,  -1816 ],  # -1.0
        [  -7849,  -1487 ],  # -0.8
        [  -6885,  -1141 ],  # -0.6
        [  -4088,    260 ],  # -0.5
        [   -408,   2478 ],  # -0.4
        [   2176,   4614 ],  # -0.3
        [   3050,   5423 ],  # -0.25
        [   3816,   6050 ],  # -0.2
        [   4229,   6382 ],  # -0.17
        [   4472,   6579 ],  # -0.15
        [   4680,   6751 ],  # -0.13
        [   4927,   6955 ],  # -0.1
        [   5107,   7052 ],  # -0.07
        [   5185,   7056 ],  # -0.05
        [   5354,   7045 ],  # 0.0
        [   5513,   7025 ],  # 0.05
        [   5663,   7003 ],  # 0.1
        [   5948,   6957 ],  # 0.2
        [   6214,   6917 ],  # 0.3
        [   6696,   6857 ],  # 0.5
        [   7144,   6766 ],  # 0.8
        [   7327,   6732 ],  # 1.0
        [   7699,   6755 ],  # 1.5
        [   8113,   6906 ],  # 2.0
    ]
).normalized(10000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-20coupling-3_1_winding-5e10-drive-
fwd_40_600um_20_3_1_5e10 = Piecewise(
    [
        [ -11623,  -2886 ],  # -2.0
        [ -10847,  -2125 ],  # -1.5
        [  -9878,  -1348 ],  # -1.0
        [  -9392,   -985 ],  # -0.8
        [  -8694,   -508 ],  # -0.6
        [  -8256,   -360 ],  # -0.5
        [  -7609,   -221 ],  # -0.4
        [  -4128,   1930 ],  # -0.3
        [   -974,   4269 ],  # -0.25
        [   2142,   6978 ],  # -0.2
        [   3895,   8940 ],  # -0.15
        [   5049,  10512 ],  # -0.1
        [   5838,  11677 ],  # -0.05
        [   6192,  11925 ],  # 0.0
        [   6415,  12017 ],  # 0.05
        [   6613,  12086 ],  # 0.1
        [   6960,  12202 ],  # 0.2
        [   7282,  12311 ],  # 0.3
        [   7658,  12407 ],  # 0.5
        [   8037,  12562 ],  # 0.8
        [   8272,  12695 ],  # 1.0
        [   8790,  13075 ],  # 1.5
        [   9248,  13503 ],  # 2.0
    ]
).normalized(19000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-20coupling-3_1_winding-1e11-drive-
fwd_40_600um_20_3_1_1e11 = Piecewise(
    [
        [ -14284,  -3414 ],  # -2.0
        [ -13332,  -2406 ],  # -1.5
        [ -12167,  -1233 ],  # -1.0
        [ -11552,   -655 ],  # -0.8
        [ -10820,    -93 ],  # -0.6
        [ -10370,    199 ],  # -0.5
        [  -9825,    434 ],  # -0.4
        [  -9043,    844 ],  # -0.3
        [  -8573,   1101 ],  # -0.25
        [  -7905,   1565 ],  # -0.2
        [  -4197,   4541 ],  # -0.15
        [   1886,   9344 ],  # -0.1
        [   5011,  11888 ],  # -0.05
        [   6604,  12899 ],  # 0.0
        [   7358,  13302 ],  # 0.05
        [   7932,  13602 ],  # 0.1
        [   8423,  13730 ],  # 0.2
        [   8645,  13888 ],  # 0.3
        [   8942,  13962 ],  # 0.5
        [   9318,  14166 ],  # 0.8
        [   9554,  14252 ],  # 1.0
        [  10166,  14547 ],  # 1.5
        [  10774,  14789 ],  # 2.0
    ]
).normalized(15000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-20coupling-3_1_winding-2e11-drive-
fwd_40_600um_20_3_1_2e11 = Piecewise(
    [
        [ -17130,  -2396 ],  # -2.0
        [ -16555,  -1969 ],  # -1.5
        [ -15189,  -1044 ],  # -1.0
        [ -14486,   -584 ],  # -0.8
        [ -13641,    -20 ],  # -0.6
        [ -13147,    302 ],  # -0.5
        [ -12539,    723 ],  # -0.4
        [ -11816,   1218 ],  # -0.3
        [ -11388,   1501 ],  # -0.25
        [ -10882,   1826 ],  # -0.2
        [ -10095,   2383 ],  # -0.15
        [  -8747,   3406 ],  # -0.1
        [    747,   8265 ],  # -0.05
        [   6051,   9708 ],  # 0.0
        [   8022,  10412 ],  # 0.05
        [   8421,  10382 ],  # 0.1
        [   8817,  10592 ],  # 0.2
        [   9116,  10658 ],  # 0.3
        [   9620,  10803 ],  # 0.5
        [  10324,  11056 ],  # 0.8
        [  10785,  11303 ],  # 1.0
        [  11938,  11738 ],  # 1.5
        [  13090,  12062 ],  # 2.0
    ]
).normalized(15000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-12coupling-5_1_winding-5e10-drive-
fwd_40_600um_12_5_1_5e10 = Piecewise(
    [
        [ -11347,  -6303 ],  # -2.0
        [ -10626,  -5304 ],  # -1.5
        [  -9680,  -4520 ],  # -1.0
        [  -9156,  -4477 ],  # -0.8
        [  -8432,  -4378 ],  # -0.6
        [  -8023,  -4242 ],  # -0.5
        [  -7423,  -4024 ],  # -0.4
        [  -4149,  -2537 ],  # -0.3
        [  -1033,   -850 ],  # -0.25
        [   1286,    777 ],  # -0.2
        [   2067,   1524 ],  # -0.17
        [   2496,   2006 ],  # -0.15
        [   2874,   2482 ],  # -0.13
        [   3338,   3142 ],  # -0.1
        [   3686,   3723 ],  # -0.07
        [   3865,   4057 ],  # -0.05
        [   4147,   4400 ],  # 0.0
        [   4326,   4538 ],  # 0.05
        [   4486,   4651 ],  # 0.1
        [   4891,   4972 ],  # 0.2
        [   5437,   5353 ],  # 0.3
        [   5897,   5559 ],  # 0.5
        [   6370,   5741 ],  # 0.8
        [   6616,   5858 ],  # 1.0
        [   7278,   6265 ],  # 1.5
        [   7937,   6759 ],  # 2.0
    ]
).normalized(12000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-12coupling-5_1_winding-1e11-drive-
fwd_40_600um_12_5_1_1e11 = Piecewise(
    [
        [ -14026,  -4901 ],  # -2.0
        [ -13014,  -4368 ],  # -1.5
        [ -11814,  -3658 ],  # -1.0
        [ -11192,  -3265 ],  # -0.8
        [ -10456,  -2801 ],  # -0.6
        [ -10025,  -2587 ],  # -0.5
        [  -9522,  -2256 ],  # -0.4
        [  -8805,  -1836 ],  # -0.3
        [  -8362,  -1556 ],  # -0.25
        [  -7702,  -1105 ],  # -0.2
        [  -4268,   1258 ],  # -0.15
        [    315,   4617 ],  # -0.1
        [   2810,   7246 ],  # -0.05
        [   4359,   9016 ],  # 0.0
        [   5225,   9707 ],  # 0.05
        [   5981,  10366 ],  # 0.1
        [   6782,  10836 ],  # 0.2
        [   7084,  11037 ],  # 0.3
        [   7501,  11328 ],  # 0.5
        [   8045,  11716 ],  # 0.8
        [   8395,  12070 ],  # 1.0
        [   9248,  12388 ],  # 1.5
        [  10140,  13087 ],  # 2.0
    ]
).normalized(14000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-12coupling-5_1_winding-2e11-drive-
fwd_40_600um_12_5_1_2e11 = Piecewise(
    [
        [ -17015,  -1518 ],  # -2.0
        [ -16471,  -1243 ],  # -1.5
        [ -15036,   -499 ],  # -1.0
        [ -14316,   -134 ],  # -0.8
        [ -13477,    290 ],  # -0.6
        [ -12998,    524 ],  # -0.5
        [ -12426,    801 ],  # -0.4
        [ -11761,   1134 ],  # -0.3
        [ -11372,   1315 ],  # -0.25
        [ -10905,   1547 ],  # -0.2
        [ -10172,   1926 ],  # -0.15
        [  -8733,   2765 ],  # -0.1
        [  -1335,   6369 ],  # -0.05
        [   3759,   8707 ],  # 0.0
        [   6392,   9509 ],  # 0.05
        [   7090,   9733 ],  # 0.1
        [   7637,   9681 ],  # 0.2
        [   8041,   9839 ],  # 0.3
        [   8724,   9855 ],  # 0.5
        [   9635,  10142 ],  # 0.8
        [  10214,  10435 ],  # 1.0
        [  11594,  10686 ],  # 1.5
        [  12900,  10836 ],  # 2.0
    ]
).normalized(14000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-8coupling-7_1_winding-5e10-drive-
fwd_40_600um_8_7_1_5e10 = Piecewise(
    [
        [ -11191,  -8744 ],  # -2.0
        [ -10386,  -8104 ],  # -1.5
        [  -9433,  -7450 ],  # -1.0
        [  -8946,  -7292 ],  # -0.8
        [  -8265,  -7035 ],  # -0.6
        [  -7873,  -6805 ],  # -0.5
        [  -7292,  -6588 ],  # -0.4
        [  -4183,  -5572 ],  # -0.3
        [  -1326,  -4520 ],  # -0.25
        [    385,  -3634 ],  # -0.2
        [   1083,  -3152 ],  # -0.17
        [   1463,  -2837 ],  # -0.15
        [   1801,  -2531 ],  # -0.13
        [   2207,  -2095 ],  # -0.1
        [   2521,  -1693 ],  # -0.07
        [   2685,  -1451 ],  # -0.05
        [   2988,  -1084 ],  # 0.0
        [   3186,   -916 ],  # 0.05
        [   3382,   -764 ],  # 0.1
        [   3998,   -338 ],  # 0.2
        [   4755,     99 ],  # 0.3
        [   5265,    344 ],  # 0.5
        [   5782,    539 ],  # 0.8
        [   6062,    652 ],  # 1.0
        [   6868,   1020 ],  # 1.5
        [   7684,   1464 ],  # 2.0
    ]
).normalized(12000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-6coupling-9_1_winding-5e10-drive-
fwd_40_600um_6_9_1_5e10 = Piecewise(
    [
        [ -11138,  -9779 ],  # -2.0
        [ -10280,  -9491 ],  # -1.5
        [  -9298,  -9070 ],  # -1.0
        [  -8827,  -8906 ],  # -0.8
        [  -8171,  -8585 ],  # -0.6
        [  -7786,  -8410 ],  # -0.5
        [  -7209,  -8166 ],  # -0.4
        [  -4205,  -7466 ],  # -0.3
        [  -1652,  -6796 ],  # -0.25
        [    -89,  -6232 ],  # -0.2
        [    580,  -5919 ],  # -0.17
        [    940,  -5712 ],  # -0.15
        [   1260,  -5491 ],  # -0.13
        [   1644,  -5196 ],  # -0.1
        [   1941,  -4928 ],  # -0.07
        [   2099,  -4755 ],  # -0.05
        [   2412,  -4431 ],  # 0.0
        [   2629,  -4251 ],  # 0.05
        [   2853,  -4088 ],  # 0.1
        [   3581,  -3693 ],  # 0.2
        [   4449,  -3306 ],  # 0.3
        [   4986,  -3048 ],  # 0.5
        [   5509,  -2843 ],  # 0.8
        [   5803,  -2733 ],  # 1.0
        [   6678,  -2391 ],  # 1.5
        [   7561,  -1986 ],  # 2.0
    ]
).normalized(12000)

# ./stacked_cores_39xx.py ../../../../out/applications/stacked_cores/40-0.00059999997rad-6coupling-11_1_winding-5e10-drive-
fwd_40_600um_6_11_1_5e10 = Piecewise(
    [
        [ -10841, -10684 ],  # -2.0
        [ -10118, -10416 ],  # -1.5
        [  -9209, -10033 ],  # -1.0
        [  -8755,  -9809 ],  # -0.8
        [  -8113,  -9520 ],  # -0.6
        [  -7741,  -9312 ],  # -0.5
        [  -7180,  -9051 ],  # -0.4
        [  -4152,  -8422 ],  # -0.3
        [  -1328,  -7790 ],  # -0.25
        [    117,  -7317 ],  # -0.2
        [   1037,  -6894 ],  # -0.15
        [   1580,  -6541 ],  # -0.1
        [   1870,  -6257 ],  # -0.05
        [   2055,  -6046 ],  # 0.0
        [   2191,  -5921 ],  # 0.05
        [   2366,  -5783 ],  # 0.1
        [   3097,  -5430 ],  # 0.2
        [   4014,  -5084 ],  # 0.3
        [   4496,  -4826 ],  # 0.5
        [   4942,  -4641 ],  # 0.8
        [   5175,  -4562 ],  # 1.0
        [   5884,  -4342 ],  # 1.5
        [   6611,  -4071 ],  # 2.0
    ]
).normalized(11000)

for (name, curve) in [
    # ("fake step", fwd_fake_step.logically_inverted()),
    # ("fake 1.5x", fwd_fake_1_5x.logically_inverted()),
    # ("fake slope-change", fwd_fake_slope_change_before_0_5.logically_inverted()),
    # ("fake slope-change (delayed)", fwd_fake_slope_change_after_0_5.logically_inverted()),
    # ("fake slope-change (delayed, shifted)", fwd_fake_slope_change_after_0_5.shifted_x(-0.1).logically_inverted()),
    # ("fake slope-change (delayed, shifted, inv-xy)", fwd_fake_slope_change_after_0_5.shifted_x(-0.1).shifted_y(-0.2).logically_inverted_x()),
    # ("fake slope-change (delayed, flipped)", fwd_fake_slope_change_after_0_5.logically_inverted_x().logically_inverted()),
    # ("fake hill", fwd_fake_hill.logically_inverted()),
    # ("fake asymmetric hill", fwd_fake_asymmetric_hill.logically_inverted()),
    # ("fake asymmetric flats", fwd_fake_asymmetric_flats.logically_inverted()),
    # ("fake asymmetric overdrive", fwd_fake_asymmetric_overdrive.logically_inverted()),
    # ("fake asymmetric bottom out", fwd_fake_asymmetric_bottom_out.logically_inverted()),

    # ("18", fwd_18.logically_inverted()),
    # ("24 5:1 (2e10 I)", fwd_24_5_1_2e10.logically_inverted()),
    # ("24 5:1 (5e10 I)", fwd_24_5_1_5e10.logically_inverted()),
    # ("24 5:1 (8e10 I)", fwd_24_5_1_8e10.logically_inverted()),
    # ("26", fwd_26.logically_inverted()),
    # ("38 1:0 (2e10 I)", fwd_38_1_0.logically_inverted()),
    # ("38 1:0 (5e10 I)", fwd_38_1_0_5e10.logically_inverted()),
    # ("38 2:0 (2e10 I)", fwd_38_2_0.logically_inverted()),
    # ("38 2:0 (5e10 I)", fwd_38_2_0_5e10.logically_inverted()),
    # ("38 3:0 (2e10 I)", fwd_38_3_0.logically_inverted()),
    # ("38 3:0 (5e10 I)", fwd_38_3_0_5e10.logically_inverted()),
    # ("38 4:0 (2e10 I)", fwd_38_4_0.logically_inverted()),
    # ("38 4:0 (5e10 I)", fwd_38_4_0_5e10.logically_inverted()),
    # ("39 2:0 (2e10 I)", inv_39_2_0_2e10),
    # ("39 2:0 (5e10 I)", inv_39_2_0_5e10),
    # ("39 2:0 (8e10 I)", inv_39_2_0_8e10),
    # ("39 2:0 (1e11 I)", inv_39_2_0_1e11),
    # ("39 2:0 (15e10 I)", inv_39_2_0_15e10),

    # ("40 12x  3:1 (400um, 5e10 I)", fwd_40_12_3_1_5e10.logically_inverted()),
    # ("40  6x  3:1 (400um, 5e10 I)", fwd_40_6_3_1_5e10.logically_inverted()),
    # ("40  6x  7:1 (400um, 5e10 I)", fwd_40_6_7_1_5e10.logically_inverted()),
    ("40 20x  3:1 (600um, 3e10 I)", fwd_40_600um_20_3_1_3e10.logically_inverted()),
    ("40 20x  3:1 (600um, 5e10 I)", fwd_40_600um_20_3_1_5e10.logically_inverted()),
    # # ("40 20x  3:1 (600um, 5e10 I)-flipped-xy", fwd_40_600um_20_3_1_5e10.logically_inverted_x()),
    # # ("40 20x  3:1 (600um, 5e10 I)-flipped", fwd_40_600um_20_3_1_5e10.logically_inverted_x()),
    # # ("40 20x  3:1 (600um, 5e10 I)-shifted", fwd_40_600um_20_3_1_5e10.shifted_x(-0.10).logically_inverted()),
    # ("40 20x  3:1 (600um, 1e11 I)", fwd_40_600um_20_3_1_1e11.logically_inverted()),
    # ("40 20x  3:1 (600um, 2e11 I)", fwd_40_600um_20_3_1_2e11.logically_inverted()),
    ("40 12x  5:1 (600um, 5e10 I)", fwd_40_600um_12_5_1_5e10.logically_inverted()),
    # ("40 12x  5:1 (600um, 1e11 I)", fwd_40_600um_12_5_1_1e11.logically_inverted()),
    # ("40 12x  5:1 (600um, 2e11 I)", fwd_40_600um_12_5_1_2e11.logically_inverted()),
    ("40  8x  7:1 (600um, 5e10 I)", fwd_40_600um_8_7_1_5e10.logically_inverted()),
    ("40  6x  9:1 (600um, 5e10 I)", fwd_40_600um_6_9_1_5e10.logically_inverted()),
    # ("40  6x 11:1 (600um, 5e10 I)", fwd_40_600um_6_11_1_5e10.logically_inverted()),
]:
    curve.plot(title = f"{name} mapping")
    curve.logically_inverted().plot_slope(title = f"{name} slope")
    curve.plot_equilibrium(title = f"{name} equilibrium")
    # curve.plot_integral(title = f"{name} integrated")
